@@ -1,4 +1,5 @@
 import argparse
+import re
 import time
 from collections import defaultdict
 from fileinput import input
@@ -12,8 +13,9 @@ from ome_zarr.io import parse_url
 from ome_zarr.reader import Multiscales, Node
 from ome_zarr.scale import Scaler
 from ome_zarr.types import JSONDict
-from omero.model import MaskI
+from omero.model import MaskI, PolygonI
 from omero.rtypes import unwrap
+from skimage.draw import polygon as sk_polygon
 from zarr.convenience import open as zarr_open
 
 from .util import print_status
@@ -35,10 +37,19 @@ MASK_DTYPE_SIZE: Dict[int, np.dtype] = {
     64: np.int64,
 }
 
+OME_MODEL_POINT_LIST_RE = re.compile(r"([\d.]+),([\d.]+)")
 
-def plate_masks_to_zarr(
-    plate: omero.gateway.PlateWrapper, args: argparse.Namespace
+SHAPE_TYPES = {"Mask": MaskI, "Polygon": PolygonI}
+
+
+def plate_shapes_to_zarr(
+    plate: omero.gateway.PlateWrapper, shape_types: List[str], args: argparse.Namespace
 ) -> None:
+    """
+    Export shapes of type "Mask" or "Polygon" on a Plate to OME-Zarr labels
+
+    @param shape_types      e.g. ["Mask", "Polygon"]
+    """
 
     gs = plate.getGridSize()
     n_rows = gs["rows"]
@@ -64,7 +75,7 @@ def plate_masks_to_zarr(
                 img = ws.getImage()
                 plate_path = f"{row}/{col}/{field_name}"
                 saver.set_image(img, plate_path)
-                masks = get_masks(img)
+                masks = get_shapes(img, shape_types)
                 if masks:
                     if args.label_map:
                         label_map = get_label_map(masks, args.label_map)
@@ -87,12 +98,20 @@ def get_label_map(masks: Dict, label_map_arg: str) -> Dict:
             line = line.strip()
             sid, name, roi = line.split(",")
             label_map[name].append(roi_map[int(roi)])
-    except Exception as e:
+    except (KeyError, ValueError) as e:
+        # KeyError: if a roi is missing from the roi_map
+        # ValueError: if there aren't enough separators
         print(f"Error parsing {label_map_arg}: {e}")
     return label_map
 
 
-def get_masks(image: omero.gateway.ImageWrapper) -> Dict:
+def get_shapes(image: omero.gateway.ImageWrapper, shape_types: List[str]) -> Dict:
+
+    shape_classes = []
+    for klass in shape_types:
+        if klass in SHAPE_TYPES:
+            shape_classes.append(SHAPE_TYPES[klass])
+
     conn = image._conn
     roi_service = conn.getRoiService()
     result = roi_service.findByImage(image.id, None, {"omero.group": "-1"})
@@ -102,7 +121,7 @@ def get_masks(image: omero.gateway.ImageWrapper) -> Dict:
     for roi in result.rois:
         mask_shapes = []
         for s in roi.copyShapes():
-            if isinstance(s, MaskI):
+            if isinstance(s, tuple(shape_classes)):
                 mask_shapes.append(s)
 
         if len(mask_shapes) > 0:
@@ -113,11 +132,16 @@ def get_masks(image: omero.gateway.ImageWrapper) -> Dict:
     return masks
 
 
-def image_masks_to_zarr(
-    image: omero.gateway.ImageWrapper, args: argparse.Namespace
+def image_shapes_to_zarr(
+    image: omero.gateway.ImageWrapper, shape_types: List[str], args: argparse.Namespace
 ) -> None:
+    """
+    Export shapes of type "Mask" or "Polygon" on an Image to OME-Zarr labels
 
-    masks = get_masks(image)
+    @param shape_types      e.g. ["Mask", "Polygon"]
+    """
+
+    masks = get_shapes(image, shape_types)
 
     dtype = MASK_DTYPE_SIZE[int(args.label_bits)]
 
@@ -272,21 +296,28 @@ class MaskSaver:
             assert False, "6d has been removed"
 
         # Create and store binary data
-        labels, fill_colors = self.masks_to_labels(
+        labels, fill_colors, properties = self.masks_to_labels(
             masks, mask_shape, ignored_dimensions, check_overlaps=True,
         )
         scaler = Scaler(max_layer=input_pyramid_levels)
         label_pyramid = scaler.nearest(labels)
-        pyramid_grp = out_labels.create_group(name)
+        pyramid_grp = out_labels.require_group(name)
+
         write_multiscale(label_pyramid, pyramid_grp)  # TODO: dtype, chunks, overwite
 
         # Specify and store metadata
         image_label_colors: List[JSONDict] = []
+        label_properties: List[JSONDict] = []
         image_label = {
-            "version": "0.1",
+            "version": "0.2",
             "colors": image_label_colors,
+            "properties": label_properties,
             "source": {"image": source_image_link},
         }
+        if properties:
+            for label_value, props_dict in sorted(properties.items()):
+                new_dict: Dict = {"label-value": label_value, **props_dict}
+                label_properties.append(new_dict)
         if fill_colors:
             for label_value, rgba_int in sorted(fill_colors.items()):
                 image_label_colors.append(
@@ -306,6 +337,13 @@ class MaskSaver:
             attrs["labels"] = [name]
         out_labels.attrs.update(attrs)
 
+    def shape_to_binim_yx(
+        self, shape: omero.model.Shape
+    ) -> Tuple[np.ndarray, Tuple[int, ...]]:
+        if isinstance(shape, MaskI):
+            return self._mask_to_binim_yx(shape)
+        return self._polygon_to_binim_yx(shape)
+
     def _mask_to_binim_yx(
         self, mask: omero.model.Shape
     ) -> Tuple[np.ndarray, Tuple[int, ...]]:
@@ -338,6 +376,37 @@ class MaskSaver:
 
         return binarray, (t, c, z, y, x, h, w)
 
+    def _polygon_to_binim_yx(
+        self, polygon: omero.model.Shape
+    ) -> Tuple[np.ndarray, Tuple[int, ...]]:
+
+        t = unwrap(polygon.theT)
+        c = unwrap(polygon.theC)
+        z = unwrap(polygon.theZ)
+
+        # "10,20, 50,150, 200,200, 250,75"
+        points = unwrap(polygon.points).strip()
+        coords = OME_MODEL_POINT_LIST_RE.findall(points)
+        x_coords = np.array([int(round(float(xy[0]))) for xy in coords])
+        y_coords = np.array([int(round(float(xy[1]))) for xy in coords])
+
+        # bounding box of polygon
+        x = x_coords.min()
+        y = y_coords.min()
+        w = x_coords.max() - x
+        h = y_coords.max() - y
+
+        img = np.zeros((h, w), dtype=self.dtype)
+
+        # coords *within* bounding box
+        x_coords = x_coords - x
+        y_coords = y_coords - y
+
+        pixels = sk_polygon(y_coords, x_coords, img.shape)
+        img[pixels] = 1
+
+        return img, (t, c, z, y, x, h, w)
+
     def _get_indices(
         self, ignored_dimensions: Set[str], d: str, d_value: int, d_size: int
     ) -> List[int]:
@@ -356,16 +425,20 @@ class MaskSaver:
         mask_shape: Tuple[int, ...],
         ignored_dimensions: Set[str] = None,
         check_overlaps: bool = True,
-    ) -> Tuple[np.ndarray, Dict[int, str]]:
+    ) -> Tuple[np.ndarray, Dict[int, str], Dict[int, Dict]]:
         """
         :param masks [MaskI]: Iterable container of OMERO masks
         :param mask_shape 5-tuple: the image dimensions (T, C, Z, Y, X), taking
             into account `ignored_dimensions`
+
         :param ignored_dimensions set(char): Ignore these dimensions and set
             size to 1
+
         :param check_overlaps bool: Whether to check for overlapping masks or
             not
-        :return: Label image with size `mask_shape` as well as color metadata.
+
+        :return: Label image with size `mask_shape` as well as color metadata
+            and dict of other properties.
 
 
         TODO: Move to https://github.com/ome/omero-rois/
@@ -390,13 +463,23 @@ class MaskSaver:
             ), f"Invalid label shape: {labels.shape}, expected {mask_shape}"
 
         fillColors: Dict[int, str] = {}
+        properties: Dict[int, Dict] = {}
+
         for count, shapes in enumerate(masks):
-            # All shapes same color for each ROI
-            for mask in shapes:
-                # Unused metadata: the{ZTC}, x, y, width, height, textValue
-                if mask.fillColor:
-                    fillColors[count + 1] = unwrap(mask.fillColor)
-                binim_yx, (t, c, z, y, x, h, w) = self._mask_to_binim_yx(mask)
+            for shape in shapes:
+                # Using ROI ID allows stitching label from multiple images
+                # into a Plate and not creating duplicates from different iamges.
+                # All shapes will be the same value (color) for each ROI
+                shape_value = shape.roi.id.val
+                properties[shape_value] = {
+                    "omero:shapeId": shape.id.val,
+                    "omero:roiId": shape.roi.id.val,
+                }
+                if shape.textValue:
+                    properties[shape_value]["omero:text"] = unwrap(shape.textValue)
+                if shape.fillColor:
+                    fillColors[shape_value] = unwrap(shape.fillColor)
+                binim_yx, (t, c, z, y, x, h, w) = self.shape_to_binim_yx(shape)
                 for i_t in self._get_indices(ignored_dimensions, "T", t, size_t):
                     for i_c in self._get_indices(ignored_dimensions, "C", c, size_c):
                         for i_z in self._get_indices(
@@ -411,12 +494,12 @@ class MaskSaver:
                                 )
                             ):
                                 raise Exception(
-                                    f"Mask {count} overlaps with existing labels"
+                                    f"Mask {shape_value} overlaps with existing labels"
                                 )
                             # ADD to the array, so zeros in our binarray don't
-                            # wipe out previous masks
+                            # wipe out previous shapes
                             labels[i_t, i_c, i_z, y : (y + h), x : (x + w)] += (
-                                binim_yx * (count + 1)  # Prevent zeroing
+                                binim_yx * shape_value
                             )
 
-        return labels, fillColors
+        return labels, fillColors, properties
